//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#ifndef OBJECT_MIXINS_HH
#define OBJECT_MIXINS_HH

#include "objects.hh"
#include "actors.hh"

namespace world
{
    template <class T> 
    class CloneFloor : public virtual Floor {
    public:
        Object* clone() { return new T; }
        void dispose() { delete this; }
    };

//----------------------------------------
// Item Mixins
//----------------------------------------
    class StaticItem : public virtual Item {
    protected:
        bool actor_hit(Actor *) { return false; }
    };
  
    template <class T> 
    class CloneItem : public virtual Item {
    protected:
        Object* clone() { return new T; }
        void dispose() { delete this; }
    };

//----------------------------------------
// CloneStone
//----------------------------------------
    template <class T>
    class CloneStone : public virtual Stone {
    protected: 
        CloneStone() {}
    private:
        Object *clone() { return new T; }
        void dispose() { delete this; }
    };

//----------------------------------------
// MovableStone
//
// This mixin class tries to move the stone to a neighboring field
// when hit hard enough.
//----------------------------------------
    class MovableStone : public virtual Stone {
    protected:
        MovableStone() {}
        
        bool on_move() { return true; }

        StoneResponse actor_hit(const StoneContact &sc)
        {
            ActorInfo *ai = sc.actor->get_actorinfo();
            
            // If the speed component towards the face of the stone is large
            // enough (and actually pointing towards the stone), consider
            // moving the stone.
            Direction dir=contact_face(sc);
            if (dir!=enigma::NODIR && ai->vel * sc.normal < -4)
                MaybeMoveStone(get_pos(), reverse(dir));
            return STONE_REBOUND;
        }
    };
    
//----------------------------------------
// OnOffStone
//
// A stone that can be on and off and understands the messages "on",
// "off", and "onoff".  Whenever the "on" attribute changes,
// Stone::set_model() is invoked
//----------------------------------------
    class OnOffStone : public virtual Stone {
        virtual void notify_onoff(bool on) {}
    protected:

        OnOffStone() { set_attrib("on", 0.0); }

        bool is_on() { return int_attrib("on")==1; }

        void set_on(bool newon)
        {
            if (newon != is_on()) {
                set_attrib("on", Value(newon));
                set_model();
                notify_onoff(newon);
            }
        }

        void message(const string &m, const Value &) 
        {
            if (m == "on")
                set_on(true);
            else if (m=="off")
                set_on(false);
            else if (m=="onoff")
                set_on(!is_on());
        }
    };
}

#endif
