//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#include "display.hh"
#include "player.hh"
#include "tools.hh"
#include "sound.hh"
#include "objects.hh"
#include "actors.hh"
#include "px/tools.hh"
#include <algorithm>
#include <iostream>
#include <cassert>

using namespace player;
using namespace std;
using namespace tools;

using world::GridPos;


class PlayerInfo {
public:
    string name;
    Inventory inventory;
    Actor *main_actor;
};

typedef vector<PlayerInfo> PlayerList;


 // Local variables
namespace
{
    PlayerList players;
    int icurrent_player = 0;
    bool current_player_inhibit_pickup = false;
}

 // Inventory implementation

unsigned const Inventory::max_items = 12;

Inventory::Inventory()
    : items()
{}

Inventory::~Inventory() 
{ 
    clear();
}

void
Inventory::clear()
{
    px::delete_sequence(items.begin(), items.end()); 
    items.clear();
}

Item *
Inventory::get_item(int idx)
{
    return idx >= size() ? 0 : items[idx];
}

Item *
Inventory::yield_item(int idx)
{
    if (0 <= idx && idx < size())
    {
        Item *it = items[idx];
        items.erase(items.begin()+ idx);
        redraw();
        return it;
    }
    return 0;
}

void 
Inventory::add_item(Item *i)
{
    items.insert(items.begin(), i);
    redraw();
}

void 
Inventory::redraw()
{
    display::RedrawInventory(this);
}

void 
Inventory::rotate()
{
    if (!items.empty())
    {
        std::rotate(items.begin(), items.begin()+1, items.end());
        redraw();
    }
}

void 
Inventory::activate_first()
{
    if (!items.empty())
    {
        Item *it = items[0];
        Actor *ac = players[icurrent_player].main_actor;
        const world::ActorInfo &ai = world::GetActorInfo(ac);

        GridPos p(static_cast<int>(ai.pos[0]), 
                  static_cast<int>(ai.pos[1]));

        switch (it->activate(ac, p)) {
        case world::ITEM_DROP:
            // only drop if no item underneath and floor allows it ###
            if (world::GetItem(p) == 0) {
                items.erase(items.begin());
                redraw();
                world::SetItem(p, it);
                it->on_drop(ac);
            }
            break;
        case world::ITEM_KILL:
            delete it;
            items.erase(items.begin());
            redraw();
            break;
        case world::ITEM_KEEP:
            break;
        }

    }
}

Item * 
Inventory::yield_first()
{
    if (!items.empty())
    {
        Item *it = items[0];
        items.erase(items.begin());
        redraw();
        return it;
    }
    return 0;
}

 // Functions

void 
player::NewGame(int nplayers)
{
    assert(nplayers > 0);

    players.clear();
    players.resize(nplayers);


    for (int i=0; i<nplayers; ++i)
    {
        Inventory *inv = GetInventory(i);
        inv->add_item(world::MakeItem("it-extralife"));
        inv->add_item(world::MakeItem("it-extralife"));
    }
    display::SetInventory(GetInventory(0));
}

void
player::NewWorld()
{
    // Clear up the inventories of all players: keep only extra lifes.
    for (unsigned iplayer=0; iplayer<players.size(); ++iplayer)
    {
        Inventory *inv = GetInventory(iplayer);
        int nextralifes=0;
        for (int i=0; i<inv->size(); ++i)
            if (inv->get_item(i)->get_kind() == "it-extralife")
                nextralifes += 1;
        inv->clear();
        for (int i=0; i<nextralifes; ++i)
            inv->add_item(world::MakeItem("it-extralife"));
    }

}

Inventory *
player::GetInventory(int iplayer)
{
    return &players[iplayer].inventory;
}

void 
player::Suicide()
{
    SendMessage(players[icurrent_player].main_actor, "shatter");
}

void 
player::SetMainActor(int iplayer, Actor *a)
{
    assert(iplayer >= 0 && (unsigned)iplayer < players.size());
    players[iplayer].main_actor = a;
}

bool 
player::AllActorsDead()
{
    for (unsigned i=0; i<players.size(); ++i) {
        if (!players[i].main_actor->is_dead())
            return false;
    }
    return true;
}

void 
player::InhibitPickup(bool flag)
{
    current_player_inhibit_pickup = flag;
}

void 
player::PickupItem(Actor *a, GridPos p)
{
    int iplayer = a->int_attrib("player");
    if (iplayer < 0 || (unsigned)iplayer >= players.size()) {
        cerr << "PickupItem: illegal 'player' entry\n";
        return;
    }

    if (iplayer==icurrent_player && current_player_inhibit_pickup)
    {
        // do not pick up items if current player doesn't want to
        return;
    }

    Inventory &inv = players[iplayer].inventory;
    if (!inv.is_full())
    {
        Item *item = world::YieldItem(p);
        inv.add_item(item);
    }
}

void 
player::ActivateItem()
{
    Inventory &inv = players[icurrent_player].inventory;
    inv.activate_first();
}

void 
player::RotateInventory()
{
    sound::PlaySound("invrotate");
    Inventory &inv = players[icurrent_player].inventory;
    inv.rotate();
}
