//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#include "actors.hh"
#include "world.hh"
#include "objects.hh"
#include "items.hh"
#include "laser.hh"
#include "display.hh"
#include "player.hh"
#include "sound.hh"
#include "px/math.hh"
#include "px/tools.hh"
#include "px/array2.hh"
#include "lua.hh"

#include <iostream>
#include <algorithm>
#include <functional>
#include <map>
#include <numeric>
#include <cassert>

using namespace std;
using namespace world;

//----------------------------------------
// Sphere
//----------------------------------------
namespace
{
    class Sphere {
        V3 center;
        double radius;
    public:
        Sphere(const V3 &c, double r)
            : center(c), radius(r)
        {}
        bool contains(const V3 &p) const {
            return length(center-p) < radius;
        }
        bool intersection_point(const Sphere &s, V3 &ip) {
            double rr = radius + s.radius;
            ip = center + normalize(s.center-center)*rr/2;
            return (length(center - s.center) <= rr);
        }
    };
}

//----------------------------------------
// Field
//----------------------------------------
namespace
{
    class Field {
    public:
	Field();
	~Field();

        Floor *floor;
        Item *item;
        Stone *stone;
    };
}


Field::Field()
{
    floor=0;
    item=0;
    stone=0;
}

Field::~Field()
{
    DisposeObject(floor);
    DisposeObject(item);
    DisposeObject(stone);
}

/*
 * This class implements the "force field" that accelerates
 * objects when the mouse is moved.  Only objects that have
 * the "mouseforce" attribute set are affected by this force
 * field.
 */
namespace
{
    class MouseForce : public ForceField {
    public:
	MouseForce() {maxforce = 150;}

	void set_force(V3 f) { force=f; limit_force(); }
	void add_force(V3 f) { force+=f; limit_force(); }
	
	V3 get_force(Actor *a, V3 x, V3 v, double time) 
        {
            if (a->is_flying() || a->is_dead())
                return V3();

            double damping=lua::get_num("MouseDamping");
            if (a->int_attrib("mouseforce"))
                return force*exp(-time*damping);
            else
                return V3();
	}

	void tick(double dtime) {
            double damping=lua::get_num("MouseDamping");
            force *= exp(-dtime*damping);

            if (length(force) < 0.01)
                force = V3();
	}
    private:
	void limit_force() {
 	    if (length(force) > maxforce)
 		force *= maxforce/length(force);
	}
	
	double maxforce;	// maximum force possible
	V3 force;
    };
}

namespace
{
    typedef px::Array2<Field> FieldArray;
    typedef std::vector<Actor*> ActorList;

    typedef vector<ForceField*> ForceList;
    typedef vector<StoneContact> StoneContactList;
}

StoneContact::StoneContact(Actor *a, Stone *s,
                           const V3 &cp, const V3 &n)
    : actor(a), stone(s)
    , response(STONE_REBOUND)
    , contact_point(cp), normal(n)
    , is_collision(false)
{}

DirectionBits 
world::contact_faces(const StoneContact &sc)
{
    using namespace enigma;

    int dirs = NODIRBIT;

    if (sc.normal*V3(-1,0,0) > 0)
        dirs |= WESTBIT;
    else if (sc.normal*V3(+1,0,0) > 0)
        dirs |= EASTBIT;
    if (sc.normal*V3(0,-1,0) > 0)
        dirs |= NORTHBIT;
    else if (sc.normal*V3(0,+1,0) > 0)
        dirs |= SOUTHBIT;

    return DirectionBits(dirs);
}

Direction
world::contact_face(const StoneContact &sc)
{
    using namespace enigma;
    if (sc.normal == V3(-1,0,0)) 
        return WEST;
    else if (sc.normal == V3(1,0,0)) 
        return EAST;
    else if (sc.normal == V3(0,-1,0))
        return NORTH;
    else if (sc.normal == V3(0,1,0))
        return SOUTH;
    else 
        return NODIR;
}

//======================================================================
// VARIABLES
//======================================================================

namespace
{
    map<string, Object *> named_objects;
    ForceList forces;
    MouseForce *mouseforce;

    FieldArray *fields;         // the grid of static objects
    ActorList actorlist;           // list of movable, dynamic objects
}

tools::Timer world::g_timer;


static bool
level_contains(GridPos p)
{
    return (p.x >= 0 && p.y >= 0 &&
            p.x < int(fields->width()) && p.y < int(fields->height()));
}


//======================================================================
// PHYSICS SIMULATION
//======================================================================

// Calculate the total acceleration on an actor A at time TIME.  The
// actor's current position X and velocity V are also passed.  Note
// that the position and velocity entries in ActorInfo will be updated
// only after a *successful* time step, so they cannot be used here.
static V3
get_accel (Actor *a, const V3 & x, const V3 & v, double time)
{
    V3 f;
    GridPos p((int)x[0], (int)x[1]);

    if (!a->is_flying())
    {
        if (Floor *floor = GetFloor(p))
        {
            f += mouseforce->get_force(a,x,v, time) * floor->mousefactor();
            if (length(v) > 0.01)
                f -= floor->friction() * normalize(v);
            f += floor->get_force(a);
        }

        if (Item *item = GetItem(p))
        {
            f += item->get_force(a);
        }
    }

    // all other force fields
    for (ForceList::iterator i=forces.begin(); i != forces.end(); ++i)
        f += (*i)->get_force(a, x, v, time);

    return f / a->get_mass();
}


// This function performs one step in the numerical integration of an
// actor's equation of motion.  TIME ist the current absolute time and
// H the size of the integration step.
static void
advance_actor(Actor *a, double time, double h)
{
    const double MAXVEL = 35;
 
#if 1
    ActorInfo &ai = *a->get_actorinfo();
    V3 accel = get_accel(a, ai.pos, ai.vel, time);
    
    ai.pos += h*ai.vel;
    ai.vel += h*accel;
#else
    V3 a1 = get_accel(a, a.pos, a.vel, time);
    V3 p1 = a.pos + h*a.vel * 0.5;
    V3 v1 = a.vel + h*a1 * 0.5;

    a1 = get_accel(a, p1, v1, time + h*0.5);
    a.pos += h*v1;
    a.vel += h*a1;
#endif

    // limit maximum velocity
    double q = length(ai.vel) / MAXVEL;
    if (q > 1)
        ai.vel /= q;
}

static void
add_contact(StoneContactList &cl, StoneContact &sc)
{
    cl.push_back(sc);
}

static void
find_contact_with_stone(Actor *a, StoneContactList &cl, int x, int y)
{
    Stone* st = GetStone(GridPos(x, y));
    if (!st)
        return;

    const ActorInfo &ai = *a->get_actorinfo();
    double r = ai.radius;
    double ax = ai.pos[0];
    double ay = ai.pos[1];
    const double contact_e = 0.01;
    const double erad = 2.0/32; // edge radius

    if (!Sphere(V3(x+0.5,y+0.5,0), r+0.5*sqrt(2.0)).contains(ai.pos))
         return;

    V3 cp; // contact point on the stone
    V3 normal; // face normal

    bool is_collision=false;

    // contact with north or south face of the stone?
    if (ax>x+erad && ax<x+1-erad) {
        double dist = r+5;

        // south
        if (ay>y+1) {
            is_collision = ai.vel[1]<0;
            dist = ay-(y+1);
            cp = V3(ax, y+1,0);
            normal = V3(0,+1,0);
        }
        // north
        else if (ay<y) {
            is_collision = ai.vel[1]>0;
            dist = y-ay;
            cp = V3(ax, y,0);
            normal = V3(0,-1,0);
        }
        
	if (dist-r < contact_e) {
            StoneContact contact(a, st, cp, normal);
            contact.is_collision = is_collision;
            add_contact(cl, contact);
        }
    }
    // contact with west or east face of the stone?
    else if (ay>y+erad && ay<y+1-erad) {
        double dist=r+5;
        if (ax>x+1) { // east
            is_collision = ai.vel[0]<0;
            dist = ax-(x+1);
            cp=V3(x+1,ay,0);
            normal=V3(+1,0,0);
        }
        else if (ax<x) { // west
            is_collision=ai.vel[0]>0;
            dist = x-ax;
            cp=V3(x,ay,0);
            normal=V3(-1,0,0);
        }
        
	if (dist-r < contact_e)
	{
            StoneContact contact(a, st, cp, normal);
            contact.is_collision = is_collision;
            add_contact(cl, contact);
        }
    }
    // contact with any of the four corners?
    else {
        V3 corner;
        double cx[2] = {erad, -erad};
        for (int xcorner=0; xcorner<=1; xcorner++)
        {
            for (int ycorner=0; ycorner<=1; ycorner++)
            {
                corner[0] = x+xcorner+cx[xcorner];
                corner[1] = y+ycorner+cx[ycorner];

                V3 b=corner-V3(ax,ay,0);

                // If the actor is close enough to the corner and doesn't
                // move away from it, we have a collision.
                bool close_enough = (length(b)-r-erad < contact_e);
                if (close_enough)
                {
                    V3 normal = normalize(-b);
                    V3 cp = corner + normal*erad;

                    StoneContact contact(a, st, cp, normal);
                    contact.is_collision = b*ai.vel > 0;
                    add_contact(cl, contact);
                    return;
                }
            }
        }
    }
}

static void
find_stone_contacts(Actor *a, StoneContactList &cl)
{
    ActorInfo &ai = *a->get_actorinfo();
    int fieldx = static_cast<int>(ai.pos[0]);
    int fieldy = static_cast<int>(ai.pos[1]);

    // Test for collisions with the nearby stones
    for (int y=fieldy-1; y<=fieldy+1; y++)
        for (int x=fieldx-1; x<=fieldx+1; x++)
            find_contact_with_stone(a,cl,x,y);
}

static V3
handle_stone_contact(V3 normal, StoneContact &sc)
{
    Actor *a = sc.actor;
    ActorInfo *ai = a->get_actorinfo();

    if (sc.is_collision) {
        StoneResponse sr = sc.stone->actor_hit(sc);
        if (sr != STONE_PASS) {
            normal += sc.normal;
            normal /= length(normal);
            
            if (length(ai->vel) > 0.1)
            {
                AddSprite(display::SPRITE_EFFECT, sc.contact_point, "ring-anim");
                sound::PlaySound(sc.stone->collision_sound());
            }
        }
        if (sr == STONE_KILL) 
            KillStone(sc.stone->get_pos());
    }
    else {
        sc.stone->actor_contact(sc.actor);
    }
    return normal;
}

static V3 
reflect(V3 v, V3 normal)
{
    double n = length(normal);
    if (n != 0)
        normal /= n;
    return V3(v-2*(v*normal)*normal);
}

static void
handle_contacts(Actor *a)
{
    // Contacts with stones
    StoneContactList cl;
    find_stone_contacts(a, cl);
    V3 normal = accumulate(cl.begin(), cl.end(), V3(), handle_stone_contact);

    ActorInfo &ai = *a->get_actorinfo();
    ai.vel = reflect(ai.vel, normal);

    // Contacts with other actors
    unsigned s=actorlist.size();
    unsigned i;
    for (i=0; i<s && actorlist[i] != a; ++i)
        ;
    for (unsigned j=i+1; j<s; ++j)
    {
        ActorInfo &v = *actorlist[i]->get_actorinfo();
        ActorInfo &w = *actorlist[j]->get_actorinfo();

        V3 p1 = v.pos;
        V3 p2 = w.pos;
        double r1 = v.radius;
        double r2 = w.radius;

        bool contactp = length(p1-p2) < r1+r2;
        if (contactp)
        {
            V3 n = normalize(p1-p2); // normal to contact surface
            V3 v1 = n * (v.vel*n);
            V3 v2 = n * (w.vel*n);
            bool collisionp = (v2-v1)*(p2-p1) < 0;
            if (collisionp) {
                double m1 = v.mass;
                double m2 = w.mass;
                double M = m1+m2;
                
                v.vel += -v1 + (v2*(2*m2) + v1*(m1-m2)) / M;
                w.vel += -v2 + (v1*(2*m1) + v2*(m2-m1)) / M;
                sound::PlaySound("ballcollision");
            }
        }
    }
}

static void
move_actor(Actor *a)
{
//     if (Stone *st = GetStone(field))
//         st->actor_inside(a);
    a->move();
}

static void
tick_actor(Actor *a, double dtime)
{
    ActorInfo &ai = * a->get_actorinfo();
    ai.oldpos = ai.pos;
    ai.oldvel = ai.vel;

    double rest_time = dtime;
    double dt = dtime;

    handle_contacts(a);
    while (rest_time > 0)
    {
        advance_actor(a, dtime-rest_time, dt);
                
        double l = length(ai.pos-ai.oldpos);

        // time step too large? try again
        if (l > 0.04 && dt > 1/1000.) {
            dt /= 2;
            ai.pos = ai.oldpos;
            ai.vel = ai.oldvel;
            continue;
        }

        // ok, accept this step
        move_actor(a);
        handle_contacts(a);

        rest_time -= dt;
        ai.oldpos = ai.pos;
        ai.oldvel = ai.vel;

        // time step too small?
        if (l < 0.01) {
            dt *= 2;
            dt = dt > rest_time? rest_time : dt;
        }
    }
}

 // world:: global functions

void
world::Tick(double dtime)
{
    for (ActorList::iterator i=actorlist.begin(); i!=actorlist.end(); ++i) {
        if (!(*i)->get_actorinfo()->grabbed)
            tick_actor(*i, dtime);
    }
    // Tell actors about their new position
//     for (ActorList::iterator i=actorlist.begin(); i!=actorlist.end(); ++i)
//         i->actor->move(i->pos);
    
    mouseforce->tick(dtime);
    for_each(forces.begin(), forces.end(),
             bind2nd(mem_fun(&ForceField::tick), dtime));

    g_timer.tick(dtime);
    laser::RecalcLightNow();   // recalculate laser beams if necessary
}

static void
clear_world()
{
    delete fields;
    delete mouseforce;
    fields=0;
    mouseforce=0;
    px::delete_sequence(forces.begin(), forces.end());
    forces.clear();

//     for_each(actorlist.begin(), actorlist.end(), mem_fun_ref(&ActorInfo::dispose));
    actorlist.clear();
}

void
world::Create(int w, int h)
{
    clear_world();
    fields = new FieldArray(w,h);
    mouseforce = new MouseForce;
    display::NewWorld(w, h);
    player::NewWorld();
    g_timer.clear();
}

void
world::Shutdown()
{
    clear_world();
}

void
world::Load(const string &name)
{
    string filename = enigma::FindDataFile(string("levels/") + name + ".lua");
    try {
        lua::Dofile(filename);
        laser::RecalcLightNow(); // recalculate laser beams if necessary
    }
    catch (lua::Error) {
        fprintf(stderr, "Error while loading landscape '%s'\n",name.c_str());
        clear_world();
        throw;
    }
}

void
world::SetMouseForce(V3 f)
{
    mouseforce->add_force(f);
}

void
world::NameObject(Object *obj, const std::string &name)
{
    named_objects[name] = obj;
}

void 
world::UnnameObject(Object *obj)
{
    if (obj && obj->get_attrib("name")) {
        const char *name = obj->string_attrib("name");
//        named_objects.erase(name); 
        named_objects[name] = 0;
    }
}


Object *
world::GetNamedObject(const std::string &name)
{
    if (named_objects.find(name) != named_objects.end())
        return named_objects[name];
    else
        return 0;
}

void
world::AddForceField(ForceField *ff)
{
    forces.push_back(ff);
}



namespace
{
    Field &get_field(GridPos p) {
        return fields->get(p.x, p.y);
    }
}


//----------------------------------------
// Layer
//
// Changes to a layer are cached and only applied at the end of a
// tick. [### not implemented]
//----------------------------------------
template <class T>
class Layer {
    T *defaultval;
public:
    Layer(T* deflt = 0) : defaultval(deflt) {}
    T *get(GridPos p) {
        return (level_contains(p)) ? raw_get(p) : defaultval;
    }

    T *yield(GridPos p) {
        if (level_contains(p)) {
            T *x = raw_get(p);
            if (x) {
                raw_set(p, 0);
                x->removal();
            }
            return x;
        } else
            return defaultval;
    }

    void set(GridPos p, T *x) {
        if (x && level_contains(p)) {
            kill(p);
            raw_set(p, x);
            x->creation(p);
        }
        else
            dispose(x);
    }
    void kill(GridPos p) { dispose(yield(p)); }

    virtual T* raw_get(GridPos) = 0;
    virtual void raw_set(GridPos, T *) = 0;

private:
    void dispose(T *x) { DisposeObject(x); }
};

//----------------------------------------
// Floor manipulation.
//----------------------------------------

namespace
{
    class FloorLayer : public Layer<Floor> {
    private:
        Floor *raw_get(GridPos p) { return get_field(p).floor; }
        void raw_set(GridPos p, Floor *x) {get_field(p).floor = x;}
    };
    FloorLayer fl_layer;
}

void world::KillFloor(GridPos p) {fl_layer.kill(p);}
Floor *world::GetFloor(GridPos p) {return fl_layer.get(p);}
void world::SetFloor(GridPos p, Floor* st) {fl_layer.set(p,st);}

//----------------------------------------
// Stone manipulation.
//----------------------------------------

namespace
{
    class BorderStone : public Stone {
    public:
        BorderStone() : Stone("borderstone") {}
        Object *clone() { return this; }
        void dispose() {}
    };

    BorderStone borderstone;


    class StoneLayer : public Layer<Stone> {
    public:
        StoneLayer() : Layer<Stone>(&borderstone) {}
    private:
        Stone *raw_get(GridPos p) { return get_field(p).stone; }
        void raw_set(GridPos p, Stone *st) {get_field(p).stone = st;}
    };

    StoneLayer st_layer;

    void stone_change(GridPos p) {
        Stone *st = GetStone(p);
        if (Item *it = GetItem(p))
            it->stone_change(st);
        if (Floor *fl = GetFloor(p))
            fl->stone_change(st);
    }

}


void world::KillStone(GridPos p) {
    st_layer.kill(p);
    stone_change(p);
}

Stone *
world::GetStone(GridPos p) {return st_layer.get(p);}

Stone *
world::YieldStone(GridPos p)
{
    Stone *st = st_layer.yield(p);
    stone_change(p);
    return st;
}

void
world::SetStone(GridPos p, Stone* st)
{
    st_layer.set(p,st);
    stone_change(p);
}

bool
world::MaybeMoveStone(GridPos p, Direction dir)
{
    GridPos newp = move(p,dir);
    if (GetStone(newp) == 0) 
    {
        Stone *st = GetStone(p);
        if (st && st->on_move()) 
        {
            sound::PlaySound("st-move");
            SetStone(newp, YieldStone(p));
            return true;
        }
    }
    return false;
}

void 
world::SwapStones(GridPos p, GridPos newp)
{
    Stone *tmp = st_layer.yield(newp);
    if (tmp->get_kind() != "borderstone") {
        sound::PlaySound("st-move");
        SetStone(newp, st_layer.yield(p));
        SetStone(p, tmp);
    }
}

//----------------------------------------
// Item manipulation.
//----------------------------------------

namespace
{
    class ItemLayer : public Layer<Item> {
    private:
        Item *raw_get(GridPos p) { return get_field(p).item; }
        void raw_set(GridPos p, Item *x) {get_field(p).item = x;}
    };
    ItemLayer it_layer;
}

void world::KillItem(GridPos p) {it_layer.kill(p);}
Item *world::GetItem(GridPos p) {return it_layer.get(p);}
Item *world::YieldItem(GridPos p) {return it_layer.yield(p);}
void world::SetItem(GridPos p, Item* st) {it_layer.set(p,st);}

//----------------------------------------
// Actor manipulation.
//----------------------------------------

void
world::SetActor(double x, double y, Actor* a)
{
    V3 pos(x,y,0);
    actorlist.push_back(a);
    a->get_actorinfo()->pos = pos;
    a->on_creation(pos);
}

void
world::GrabActor(Actor *a)
{
    a->get_actorinfo()->grabbed = true;
}

void
world::ReleaseActor(Actor *a)
{
    a->get_actorinfo()->grabbed = false;
}

const ActorInfo &
world::GetActorInfo(Actor *a)
{
    return *a->get_actorinfo();
}

void 
world::Init()
{
    actors::Init();
    laser::Init();
    items::Init();
}
